# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Library for testing Tensorflow Transform."""


import six
import tensorflow as tf
import tensorflow_transform as tft
from tensorflow_transform import test_case
from tensorflow_transform.beam import impl as beam_impl
from tensorflow_transform.beam.tft_beam_io import transform_fn_io

parameters = test_case.parameters


class TransformTestCase(test_case.TransformTestCase):
  """Base test class for testing tf-transform preprocessing functions."""

  def assertAnalyzeAndTransformResults(self,
                                       input_data,
                                       input_metadata,
                                       preprocessing_fn,
                                       expected_data=None,
                                       expected_metadata=None,
                                       only_check_core_metadata=False,
                                       expected_vocab_file_contents=None,
                                       expected_asset_file_contents=None,
                                       test_data=None,
                                       desired_batch_size=None):
    """Assert that input data and metadata is transformed as expected.

    This methods asserts transformed data and transformed metadata match
    with expected_data and expected_metadata.

    Args:
      input_data: A sequence of dicts whose values are
          either strings, lists of strings, numeric types or a pair of those.
      input_metadata: DatasetMetadata describing input_data.
      preprocessing_fn: A function taking a dict of tensors and returning
          a dict of tensors.
      expected_data: (optional) A dataset with the same type constraints as
          input_data, but representing the output after transformation.
          If supplied, transformed data is asserted to be equal.
      expected_metadata: (optional) DatasetMetadata describing the transformed
          data. If supplied, transformed metadata is asserted to be equal.
      only_check_core_metadata: A boolean to indicate if all elements in
          the transformed metadata is asserted to be equal to expected metadata.
          If True, only transformed feature names, dtypes and representations
          are asserted.
      expected_vocab_file_contents: (optional) A dictionary from vocab filenames
          to their expected content as a list of text lines.  Values should be
          the expected result of calling f.readlines() on the given asset files.
      expected_asset_file_contents: deprecated.  Use
          expected_vocab_file_contents.
      test_data: (optional) If this is provided then instead of calling
          AnalyzeAndTransformDataset with input_data, this function will call
          AnalyzeDataset with input_data and TransformDataset with test_data.
          Note that this is the case even if input_data and test_data are equal.
          test_data should also conform to input_metadata.
      desired_batch_size: (optional) A batch size to batch elements by. If not
          provided, a batch size will be computed automatically.
    Raises:
      AssertionError: if the expected data does not match the results of
          transforming input_data according to preprocessing_fn, or
          (if provided) if the expected metadata does not match.
      ValueError: if expected_vocab_file_contents and
          expected_asset_file_contents are both set.
    """
    if (expected_vocab_file_contents is not None and
        expected_asset_file_contents is not None):
      raise ValueError('only one of expected_asset_file_contents and '
                       'expected_asset_file_contents should be set')
    elif expected_asset_file_contents is not None:
      tf.logging.warn('expected_asset_file_contents is deprecated, use '
                      'expected_vocab_file_contents')

    expected_vocab_file_contents = (
        expected_vocab_file_contents or expected_asset_file_contents or {})
    del expected_asset_file_contents

    # Note: we don't separately test AnalyzeDataset and TransformDataset as
    # AnalyzeAndTransformDataset currently simply composes these two
    # transforms.  If in future versions of the code, the implementation
    # differs, we should also run AnalyzeDataset and TransformDatset composed.
    temp_dir = self.get_temp_dir()
    with beam_impl.Context(
        temp_dir=temp_dir, desired_batch_size=desired_batch_size):
      if test_data is None:
        (transformed_data, transformed_metadata), transform_fn = (
            (input_data, input_metadata)
            | beam_impl.AnalyzeAndTransformDataset(preprocessing_fn))
      else:
        transform_fn = ((input_data, input_metadata)
                        | beam_impl.AnalyzeDataset(preprocessing_fn))
        transformed_data, transformed_metadata = (
            ((test_data, input_metadata), transform_fn)
            | beam_impl.TransformDataset())

      # Write transform_fn so we can test its assets
      if expected_vocab_file_contents:
        _ = transform_fn | transform_fn_io.WriteTransformFn(temp_dir)

    if expected_data is not None:
      self.assertDataCloseOrEqual(expected_data, transformed_data)

    if expected_metadata:
      # Now that the pipeline has run, transformed_metadata.deferred_metadata
      # should be a list containing a single DatasetMetadata with the full
      # metadata.
      assert len(transformed_metadata.deferred_metadata) == 1
      transformed_metadata = transformed_metadata.deferred_metadata[0]

      if only_check_core_metadata:
        # preprocessing_fn may add metadata to column schema only relevant to
        # internal implementation such as vocabulary_file. As such, only check
        # feature names, dtypes and representations are as expected.
        self.assertSameElements(
            transformed_metadata.schema.column_schemas.keys(),
            expected_metadata.schema.column_schemas.keys())
        for k, v in transformed_metadata.schema.column_schemas.iteritems():
          expected_schema = expected_metadata.schema.column_schemas[k]
          self.assertEqual(expected_schema.representation, v.representation,
                           "representation doesn't match for feature '%s'" % k)
          self.assertEqual(expected_schema.domain.dtype, v.domain.dtype,
                           "dtype doesn't match for feature '%s'" % k)
      else:
        # Check the entire DatasetMetadata is as expected.
        # Use extra assertEqual for schemas, since full metadata assertEqual
        # error message is not conducive to debugging.
        self.assertEqual(expected_metadata.schema.column_schemas,
                         transformed_metadata.schema.column_schemas)
        self.assertEqual(expected_metadata, transformed_metadata)

    tf_transform_output = tft.TFTransformOutput(temp_dir)
    for filename, file_contents in six.iteritems(expected_vocab_file_contents):
      full_filename = tf_transform_output.vocabulary_file_by_name(filename)
      with tf.gfile.Open(full_filename) as f:
        self.assertEqual(f.readlines(), file_contents)
